package org.eclipse.jdt.internal.compiler.apt.model;

package .tmp;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ElementVisitor;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.apt.model.AnnotationMemberValue;
import org.eclipse.jdt.internal.compiler.apt.model.ElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.Factory;
import org.eclipse.jdt.internal.compiler.apt.model.NameImpl;
import org.eclipse.jdt.internal.compiler.apt.model.TypeElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.VariableElementImpl;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationHolder;
import org.eclipse.jdt.internal.compiler.lookup.AptBinaryLocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;

public class ExecutableElementImpl extends ElementImpl implements ExecutableElement {
  private Name _name = null;
  
  ExecutableElementImpl(BaseProcessingEnvImpl env, MethodBinding binding) {
    super(env, (Binding)binding);
  }
  
  public <R, P> R accept(ElementVisitor<R, P> v, P p) {
    return v.visitExecutable(this, p);
  }
  
  protected AnnotationBinding[] getAnnotationBindings() {
    return ((MethodBinding)this._binding).getAnnotations();
  }
  
  public AnnotationValue getDefaultValue() {
    MethodBinding binding = (MethodBinding)this._binding;
    Object defaultValue = binding.getDefaultValue();
    if (defaultValue != null)
      return (AnnotationValue)new AnnotationMemberValue(this._env, defaultValue, binding); 
    return null;
  }
  
  public List<? extends Element> getEnclosedElements() {
    return Collections.emptyList();
  }
  
  public Element getEnclosingElement() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (binding.declaringClass == null)
      return null; 
    return this._env.getFactory().newElement((Binding)binding.declaringClass);
  }
  
  public String getFileName() {
    ReferenceBinding dc = ((MethodBinding)this._binding).declaringClass;
    char[] name = dc.getFileName();
    if (name == null)
      return null; 
    return new String(name);
  }
  
  public ElementKind getKind() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (binding.isConstructor())
      return ElementKind.CONSTRUCTOR; 
    if (CharOperation.equals(binding.selector, TypeConstants.CLINIT))
      return ElementKind.STATIC_INIT; 
    if (CharOperation.equals(binding.selector, TypeConstants.INIT))
      return ElementKind.INSTANCE_INIT; 
    return ElementKind.METHOD;
  }
  
  public Set<Modifier> getModifiers() {
    MethodBinding binding = (MethodBinding)this._binding;
    return Factory.getModifiers(binding.modifiers, getKind());
  }
  
  PackageElement getPackage() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (binding.declaringClass == null)
      return null; 
    return this._env.getFactory().newPackageElement(binding.declaringClass.fPackage);
  }
  
  public List<? extends VariableElement> getParameters() {
    MethodBinding binding = (MethodBinding)this._binding;
    int length = (binding.parameters == null) ? 0 : binding.parameters.length;
    if (length != 0) {
      AbstractMethodDeclaration methodDeclaration = binding.sourceMethod();
      List<VariableElement> params = new ArrayList<>(length);
      if (methodDeclaration != null) {
        byte b;
        int i;
        Argument[] arrayOfArgument;
        for (i = (arrayOfArgument = methodDeclaration.arguments).length, b = 0; b < i; ) {
          Argument argument = arrayOfArgument[b];
          VariableElementImpl variableElementImpl = new VariableElementImpl(this._env, (VariableBinding)argument.binding);
          params.add(variableElementImpl);
          b++;
        } 
      } else {
        AnnotationBinding[][] parameterAnnotationBindings = null;
        AnnotationHolder annotationHolder = binding.declaringClass.retrieveAnnotationHolder((Binding)binding, false);
        if (annotationHolder != null)
          parameterAnnotationBindings = annotationHolder.getParameterAnnotations(); 
        int i = 0;
        byte b;
        int j;
        TypeBinding[] arrayOfTypeBinding;
        for (j = (arrayOfTypeBinding = binding.parameters).length, b = 0; b < j; ) {
          TypeBinding typeBinding = arrayOfTypeBinding[b];
          char[] name = (binding.parameterNames.length > i) ? binding.parameterNames[i] : null;
          if (name == null) {
            StringBuilder builder = new StringBuilder("arg");
            builder.append(i);
            name = String.valueOf(builder).toCharArray();
          } 
          VariableElementImpl variableElementImpl = new VariableElementImpl(this._env, 
              (VariableBinding)new AptBinaryLocalVariableBinding(
                name, 
                typeBinding, 
                0, 
                (parameterAnnotationBindings != null) ? parameterAnnotationBindings[i] : null, 
                binding));
          params.add(variableElementImpl);
          i++;
          b++;
        } 
      } 
      return Collections.unmodifiableList(params);
    } 
    return Collections.emptyList();
  }
  
  public TypeMirror getReturnType() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (binding.returnType == null)
      return null; 
    return this._env.getFactory().newTypeMirror((Binding)binding.returnType);
  }
  
  public Name getSimpleName() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (this._name == null)
      this._name = (Name)new NameImpl(binding.selector); 
    return this._name;
  }
  
  public List<? extends TypeMirror> getThrownTypes() {
    MethodBinding binding = (MethodBinding)this._binding;
    if (binding.thrownExceptions.length == 0)
      return Collections.emptyList(); 
    List<TypeMirror> list = new ArrayList<>(binding.thrownExceptions.length);
    byte b;
    int i;
    ReferenceBinding[] arrayOfReferenceBinding;
    for (i = (arrayOfReferenceBinding = binding.thrownExceptions).length, b = 0; b < i; ) {
      ReferenceBinding exception = arrayOfReferenceBinding[b];
      list.add(this._env.getFactory().newTypeMirror((Binding)exception));
      b++;
    } 
    return list;
  }
  
  public List<? extends TypeParameterElement> getTypeParameters() {
    MethodBinding binding = (MethodBinding)this._binding;
    TypeVariableBinding[] variables = binding.typeVariables();
    if (variables.length == 0)
      return Collections.emptyList(); 
    List<TypeParameterElement> params = new ArrayList<>(variables.length);
    byte b;
    int i;
    TypeVariableBinding[] arrayOfTypeVariableBinding1;
    for (i = (arrayOfTypeVariableBinding1 = variables).length, b = 0; b < i; ) {
      TypeVariableBinding variable = arrayOfTypeVariableBinding1[b];
      params.add(this._env.getFactory().newTypeParameterElement(variable, this));
      b++;
    } 
    return Collections.unmodifiableList(params);
  }
  
  public boolean hides(Element hidden) {
    if (!(hidden instanceof org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl))
      return false; 
    MethodBinding hiderBinding = (MethodBinding)this._binding;
    MethodBinding hiddenBinding = (MethodBinding)((org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl)hidden)._binding;
    if (hiderBinding == hiddenBinding)
      return false; 
    if (hiddenBinding.isPrivate())
      return false; 
    if (!hiderBinding.isStatic() || !hiddenBinding.isStatic())
      return false; 
    if (!CharOperation.equals(hiddenBinding.selector, hiderBinding.selector))
      return false; 
    if (!this._env.getLookupEnvironment().methodVerifier().isMethodSubsignature(hiderBinding, hiddenBinding))
      return false; 
    return (hiderBinding.declaringClass.findSuperTypeOriginatingFrom((TypeBinding)hiddenBinding.declaringClass) != null);
  }
  
  public boolean isVarArgs() {
    return ((MethodBinding)this._binding).isVarargs();
  }
  
  public boolean overrides(ExecutableElement overridden, TypeElement type) {
    MethodBinding overriddenBinding = (MethodBinding)((org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl)overridden)._binding;
    ReferenceBinding overriderContext = (ReferenceBinding)((TypeElementImpl)type)._binding;
    if ((MethodBinding)this._binding == overriddenBinding || 
      overriddenBinding.isStatic() || 
      overriddenBinding.isPrivate() || (
      (MethodBinding)this._binding).isStatic())
      return false; 
    char[] selector = ((MethodBinding)this._binding).selector;
    if (!CharOperation.equals(selector, overriddenBinding.selector))
      return false; 
    if (overriderContext.findSuperTypeOriginatingFrom((TypeBinding)((MethodBinding)this._binding).declaringClass) == null && 
      ((MethodBinding)this._binding).declaringClass.findSuperTypeOriginatingFrom((TypeBinding)overriderContext) == null)
      return false; 
    MethodBinding overriderBinding = new MethodBinding((MethodBinding)this._binding, overriderContext);
    if (overriderBinding.isPrivate())
      return false; 
    TypeBinding match = overriderBinding.declaringClass.findSuperTypeOriginatingFrom((TypeBinding)overriddenBinding.declaringClass);
    if (!(match instanceof ReferenceBinding))
      return false; 
    MethodBinding[] superMethods = ((ReferenceBinding)match).getMethods(selector);
    LookupEnvironment lookupEnvironment = this._env.getLookupEnvironment();
    if (lookupEnvironment == null)
      return false; 
    MethodVerifier methodVerifier = lookupEnvironment.methodVerifier();
    for (int i = 0, length = superMethods.length; i < length; i++) {
      if (superMethods[i].original() == overriddenBinding)
        return methodVerifier.doesMethodOverride(overriderBinding, superMethods[i]); 
    } 
    return false;
  }
  
  public TypeMirror getReceiverType() {
    return this._env.getFactory().getReceiverType((MethodBinding)this._binding);
  }
  
  public boolean isDefault() {
    if (this._binding != null)
      return ((MethodBinding)this._binding).isDefaultMethod(); 
    return false;
  }
}
